@@ -262,13 +262,26 @@ module Init : sig
 end
 
 module Log : sig
+  type entry = {
+    time : float;
+    label : string option;
+    level : int option;
+    log : string;
+  }
+
   (**
        Type for loggers.
     *)
   type t =
     < active : int -> bool
     ; path : Conf.path
-    ; f : 'a. int -> ('a, unit, string, unit) format4 -> 'a >
+    ; f : 'a. int -> ('a, unit, string, unit) format4 -> 'a
+    ; g :
+        'a.
+        ?colorize:(entry -> entry) ->
+        int ->
+        ('a, unit, string, unit) format4 ->
+        'a >
 
   type custom_log = { timestamp : bool; exec : string -> unit }
 
