@@ -83,24 +83,17 @@ module Conf = struct
       ?(l : links = []) ?(comments : string list = []) descr : 'a t =
     object (self)
       val kind : string option = kind
-
       val mutable descr : string = descr
-
       val mutable comments : string list = comments
-
       val mutable links : links = []
-
       val value_d : 'a option ref = ref d
-
       val value : 'a option ref = ref None
-
       val mutable validators : ('a -> bool) list = []
-
       val mutable listeners : ('a -> unit) list = []
 
       initializer
-      p self#ut;
-      List.iter (fun (s, t) -> self#plug s t) l
+        p self#ut;
+        List.iter (fun (s, t) -> self#plug s t) l
 
       method subs = List.sort compare (List.map fst links)
 
@@ -123,13 +116,9 @@ module Conf = struct
         aux [] self#ut
 
       method kind = kind
-
       method descr = descr
-
       method private set_descr new_descr = descr <- new_descr
-
       method comments = comments
-
       method private set_comments new_comments = comments <- new_comments
 
       method plug s t =
@@ -151,9 +140,7 @@ module Conf = struct
         key
 
       method ut = (self :> ut)
-
       method get_d : 'a option = !value_d
-
       method set_d (v : 'a option) : unit = value_d := v
 
       method get : 'a =
@@ -161,7 +148,7 @@ module Conf = struct
           | None -> (
               match !value_d with
                 | None -> raise (Undefined self#ut)
-                | Some v -> v )
+                | Some v -> v)
           | Some v -> v
 
       method set (v : 'a) : unit =
@@ -172,7 +159,6 @@ module Conf = struct
         List.iter (fun fn -> fn v) listeners
 
       method validate (fn : 'a -> bool) : unit = validators <- fn :: validators
-
       method on_change (fn : 'a -> unit) : unit = listeners <- fn :: listeners
     end
 
@@ -233,22 +219,22 @@ module Conf = struct
       Printf.sprintf "## %s\n" t#descr
       ^ begin
           match get_d_string t with
-          | None -> ""
-          | Some d -> Printf.sprintf "# default :%s\n" d
+            | None -> ""
+            | Some d -> Printf.sprintf "# default :%s\n" d
         end
       ^ begin
           match (t#kind, get_string t) with
-          | Some k, None -> Printf.sprintf "#%s\t%-30s\n" k prefix
-          | Some k, Some p -> Printf.sprintf "%s\t%-30s :%s\n" k prefix p
-          | _ -> ""
+            | Some k, None -> Printf.sprintf "#%s\t%-30s\n" k prefix
+            | Some k, Some p -> Printf.sprintf "%s\t%-30s :%s\n" k prefix p
+            | _ -> ""
         end
       ^ begin
           match t#comments with
-          | [] -> ""
-          | l ->
-              "# comments:\n"
-              ^ String.concat ""
-                  (List.map (fun s -> Printf.sprintf "#  %s\n" s) l)
+            | [] -> ""
+            | l ->
+                "# comments:\n"
+                ^ String.concat ""
+                    (List.map (fun s -> Printf.sprintf "#  %s\n" s) l)
         end
       ^ "\n" ^ String.concat "" subs
     in
@@ -260,14 +246,14 @@ module Conf = struct
       let subs = List.map (function s -> aux (p s) (t#path [s])) t#subs in
       begin
         match t#kind with
-        | Some k -> (
-            match (get_d_string t, get_string t) with
-              | None, None -> Printf.sprintf "#%s\t%-30s\n" k prefix
-              | Some p, None -> Printf.sprintf "#%s\t%-30s :%s\n" k prefix p
-              | Some p, Some p' when p' = p ->
-                  Printf.sprintf "#%s\t%-30s :%s\n" k prefix p
-              | _, Some p -> Printf.sprintf "%s\t%-30s :%s\n" k prefix p )
-        | _ -> ""
+          | Some k -> (
+              match (get_d_string t, get_string t) with
+                | None, None -> Printf.sprintf "#%s\t%-30s\n" k prefix
+                | Some p, None -> Printf.sprintf "#%s\t%-30s :%s\n" k prefix p
+                | Some p, Some p' when p' = p ->
+                    Printf.sprintf "#%s\t%-30s :%s\n" k prefix p
+                | _, Some p -> Printf.sprintf "%s\t%-30s :%s\n" k prefix p)
+          | _ -> ""
       end
       ^ String.concat "" subs
     in
@@ -283,7 +269,7 @@ module Conf = struct
         | "unit" -> (
             match val2 = "" with
               | false -> raise (Wrong_Conf (s, "unit expected"))
-              | true -> (as_unit st)#set () )
+              | true -> (as_unit st)#set ())
         | "int" ->
             let i =
               try int_of_string val2
@@ -311,7 +297,7 @@ module Conf = struct
         | "list" ->
             let l = Str.split list_sep_regexp val2 in
             (as_list st)#set l
-        | _ -> raise (Wrong_Conf (s, "unknown type")) )
+        | _ -> raise (Wrong_Conf (s, "unknown type")))
     else raise (Wrong_Conf (s, "syntax error"))
 
   let conf_file t s =
@@ -577,7 +563,7 @@ module Init = struct
       try
         let filename = conf_daemon_pidfile_path#get in
         Sys.remove filename
-      with _ -> () )
+      with _ -> ())
 
   exception Root_prohibited of [ `User | `Group | `Both ]
 
@@ -626,10 +612,25 @@ module Init = struct
 end
 
 module Log = struct
+  type entry = {
+    time : float;
+    label : string option;
+    level : int option;
+    log : string;
+  }
+
+  type pending_entry = { colorize : entry -> entry; entry : entry }
+
   type t =
     < active : int -> bool
     ; path : Conf.path
-    ; f : 'a. int -> ('a, unit, string, unit) format4 -> 'a >
+    ; f : 'a. int -> ('a, unit, string, unit) format4 -> 'a
+    ; g :
+        'a.
+        ?colorize:(entry -> entry) ->
+        int ->
+        ('a, unit, string, unit) format4 ->
+        'a >
 
   type custom_log = { timestamp : bool; exec : string -> unit }
 
@@ -668,25 +669,37 @@ module Log = struct
             (date.Unix.tm_year + 1900) (date.Unix.tm_mon + 1) date.Unix.tm_mday
             date.Unix.tm_hour date.Unix.tm_min date.Unix.tm_sec
 
-  let print (time, str) =
+  let message ?(show_timestamp = true) { time; label; level; log } =
+    let label =
+      match (label, level) with
+        | None, None -> ""
+        | Some l, None -> Printf.sprintf "[%s] " l
+        | None, Some d -> Printf.sprintf "[%d] " d
+        | Some l, Some d -> Printf.sprintf "[%s:%d] " l d
+    in
+    let str = label ^ log in
+    let timestamp = if show_timestamp then timestamp time ^ " " else "" in
+    Printf.sprintf "%s%s" timestamp str
+
+  let print { colorize; entry } =
     let to_stdout = conf_stdout#get in
     let to_file = !log_ch <> None in
-    let timestamp = timestamp time in
-    let message = Printf.sprintf "%s %s" timestamp str in
     begin
       match to_stdout || to_file with
-      | true ->
-          let do_stdout () = Printf.printf "%s\n%!" message in
-          let do_file () =
-            match !log_ch with
-              | None -> ()
-              | Some ch -> Printf.fprintf ch "%s\n%!" message
-          in
-          if to_stdout then do_stdout ();
-          if to_file then do_file ()
-      | false -> ()
+        | true ->
+            let do_stdout () =
+              Printf.printf "%s\n%!" (message (colorize entry))
+            in
+            let do_file () =
+              match !log_ch with
+                | None -> ()
+                | Some ch -> Printf.fprintf ch "%s\n%!" (message entry)
+            in
+            if to_stdout then do_stdout ();
+            if to_file then do_file ()
+        | false -> ()
     end;
-    let f _ x = x.exec (if x.timestamp then message else str) in
+    let f _ x = x.exec (message ~show_timestamp:x.timestamp entry) in
     Hashtbl.iter f custom_log
 
   (* Avoid interlacing logs *)
@@ -764,11 +777,9 @@ module Log = struct
     let confs = build path in
     let path_str = Conf.string_of_path path in
     object (self : t)
-      val label = fun lvl -> "[" ^ path_str ^ ":" ^ string_of_int lvl ^ "]"
-
       method path = path
 
-      method active lvl =
+      method active level =
         let rec aux l =
           match l with
             | [] -> None
@@ -777,19 +788,32 @@ module Log = struct
                   | Some i -> Some i
                   | None -> (
                       try Some (Conf.as_int t)#get
-                      with Conf.Undefined _ -> None ) )
+                      with Conf.Undefined _ -> None))
         in
-        match aux confs with None -> false | Some i -> i >= lvl
+        match aux confs with None -> false | Some i -> i >= level
 
-      method f lvl =
-        match self#active lvl with
+      method g ?(colorize = fun x -> x) level =
+        match self#active level with
           | true ->
               let time = Unix.gettimeofday () in
               Printf.ksprintf (fun s ->
                   List.iter
-                    (fun s -> proceed (time, label lvl ^ " " ^ s))
+                    (fun log ->
+                      proceed
+                        {
+                          colorize;
+                          entry =
+                            {
+                              time;
+                              label = Some path_str;
+                              level = Some level;
+                              log;
+                            };
+                        })
                     (String.split_on_char '\n' s))
           | false -> Printf.ksprintf (fun _ -> ())
+
+      method f level = self#g ?colorize:None level
     end
 
   let init () =
@@ -806,10 +830,10 @@ module Log = struct
         fun _ ->
           begin
             match !log_ch with
-            | None -> ()
-            | Some ch ->
-                log_ch := None;
-                close_out ch
+              | None -> ()
+              | Some ch ->
+                  log_ch := None;
+                  close_out ch
           end;
           log_ch := Some (open_out_gen opts log_file_perms log_file_path)
       end
@@ -818,7 +842,11 @@ module Log = struct
     (* Re-open log file on SIGUSR1 -- for logrotate *)
     if Sys.os_type <> "Win32" then
       Sys.set_signal Sys.sigusr1 (Sys.Signal_handle reopen_log);
-    print (time, ">>> LOG START");
+    print
+      {
+        colorize = (fun x -> x);
+        entry = { time; level = None; label = None; log = ">>> LOG START" };
+      };
     log_thread := Some (Thread.create log_thread_fn ())
 
   let start = Init.make ~name:"init-log-start" ~before:[Init.start] init
@@ -826,13 +854,17 @@ module Log = struct
   let close () =
     let time = Unix.gettimeofday () in
     mutexify (fun () -> log_stop := true) ();
-    proceed (time, ">>> LOG END");
+    proceed
+      {
+        colorize = (fun x -> x);
+        entry = { time; level = None; label = None; log = ">>> LOG END" };
+      };
     begin
       match !log_thread with
-      | None -> ()
-      | Some th ->
-          Condition.signal log_condition;
-          Thread.join th
+        | None -> ()
+        | Some th ->
+            Condition.signal log_condition;
+            Thread.join th
     end;
     match !log_ch with
       | None -> ()
